/*
 * Copyright 2012 Peter Lebbing <peter@digitalbrains.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef FILE_CIRCPTR_H
#define FILE_CIRCPTR_H

/*
 * Inline function to do circular buffer arithmetic.
 *
 * If bufsize is a constant and a power of two, an optimized increase is used.
 */
static inline void
circ_buf_incr_ptr(uint8_t * pointer, const uint8_t bufsize)
{
	if (__builtin_constant_p(bufsize) && (bufsize & (bufsize - 1)) == 0) {
		/* bufsize is constant and a power of 2 */
		*pointer = (*pointer + 1) & (bufsize - 1);
	} else {
		(*pointer)++;
		if (*pointer >= bufsize)
			*pointer = 0;
	}
}

/*
 * Inline function for appending the byte in argument c to a circular buffer
 * with overflow check.
 *
 * Pointer fashion: increment-after
 * 
 * Returns non-zero when the buffer is full.
 */
static inline int8_t
append_circ_buf(volatile uint8_t buf[], const uint8_t bufsize, const uint8_t c,
		uint8_t * head, const uint8_t tail)
{
	if (*head == tail) {
		/* Buffer full */
		return -1;
	}
	buf[*head] = c;
	circ_buf_incr_ptr(head, bufsize);
	return 0;
}

#endif /* ndef FILE_CIRCPTR_H */
