/*
 * Copyright 2012 Peter Lebbing <peter@digitalbrains.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <avr/io.h>
#include <stdint.h>
#include <stdio.h>
#include "common.h"
#include "circptr.h"
#include "suart.h"

/*
 * The brr holds the delay for a bitperiod, according to the formula:
 *
 * period = 9 + 3 * brr
 *
 * (period is in clockticks)
 *
 * In "small brr" mode, brr1 holds the brr, and brr_init is used to do the
 * initial delay.
 *
 * In "large brr" mode, brr2 holds bits 0 through 6 of the brr. brr1 holds
 * bits 7 through 14, plus one. The increment is because of the interaction
 * with "small brr" mode in the first part of the reception interrupt handler.
 * The initial delay is computed in the interrupt handler (brr_init is not
 * used).
 */
volatile uint8_t
suart_brr1;

/*
 * In "small brr" mode, brr2 has the bit SUART_SMALL_BRR set. In "large brr"
 * mode, this bit is cleared, and brr2 holds bits 0 through 6 of the brr.
 */
volatile uint8_t
suart_brr2;

/*
 * Initial delay for sampling the first databit in "small brr" mode. Interrupt
 * delay plus processing plus this delay loop puts the first sample 
 * 1.5 bitperiod after the falling edge of the startbit, i.e., in the middle
 * of the first databit.
 *
 * As brr1 is the delay for one bitperiod in "small brr" mode, brr_init and
 * brr1 relate as follows:
 *
 * 16+3*brr_init = 1.5*(9+3*brr1)
 * 
 * (16 is interrupt delay and processing)
 *
 * Working this out:
 *
 * 3*brr_init = 4.5*brr1 - 2.5
 * brr_init = 1.5*brr1 - 5/6
 * 
 * After some experimentation...:
 *
 * brr_init = ceil(1.5*brr1) - 1
 */
volatile uint8_t
suart_brr_init;

/*
 * Head and tail pointers.
 *
 * Pointer fashion: increment-before (!)
 */
volatile uint8_t
suart_rxh;
volatile uint8_t
suart_rxt;

volatile uint8_t
suart_buf[SUART_BUFSIZE]; /* Reception buffer */

#if SUART_AS_STDIO
static FILE mystdio = FDEV_SETUP_STREAM(suart_stdio_put, suart_stdio_get,
		_FDEV_SETUP_RW);
#endif /* SUART_AS_STDIO */

void
suart_init(uint16_t brr)
{
	if (brr < 32) {
		uint8_t brr1;

		brr1 = (uint8_t) brr;
		suart_brr2 = _BV(SUART_SMALL_BRR);
		suart_brr_init = brr1 + ((brr1 + 1) >> 1) - 1;
		suart_brr1 = brr1;
	} else {
		suart_brr_init = 1;
		suart_brr2 = brr & 0x7f;
		suart_brr1 = ((brr >> 7) & 0xff) + 1;
	}

	/* Set pin as input */
	SUART_RX_DDR &= ~(_BV(SUART_RX));
	/* Disable pull-up */
	SUART_RX_PORT &= ~(_BV(SUART_RX));
	/* Clear any pending interrupt */
	SUART_FLAG_REG = _BV(SUART_FLAG_BIT);
	/* And enable the interrupt */
	SUART_IE_REG |= _BV(SUART_IE_BIT);

	SUART_TX_PORT |= _BV(SUART_TX);
	/* Set pin as output */
	SUART_TX_DDR |= _BV(SUART_TX);

#if SUART_AS_STDIO
	stdin = &mystdio;
	stdout = &mystdio;
#endif /* SUART_AS_STDIO */

}

void
suart_enable_rx(void)
{
	if (bit_is_clear(SUART_IE_REG, SUART_IE_BIT)) {
		/* Clear and re-enable SUART RX interrupt */
		SUART_FLAG_REG = _BV(SUART_FLAG_BIT);
		SUART_IE_REG |= _BV(SUART_IE_BIT);
	}
}

uint8_t
suart_get(void)
{
	uint8_t temp_tail, c;

	temp_tail = suart_rxt;

	do {} while (bit_is_set(SUART_IE_REG, SUART_IE_BIT)
			&& temp_tail == suart_rxh);

	circ_buf_incr_ptr(&temp_tail, SUART_BUFSIZE);
	c = suart_buf[temp_tail];

	if (bit_is_clear(SUART_IE_REG, SUART_IE_BIT)) {
		/* Interrupt got disabled because the buffer was full */
		suart_rxt = temp_tail;
		/* Clear and re-enable SUART RX interrupt */
		SUART_FLAG_REG = _BV(SUART_FLAG_BIT);
		SUART_IE_REG |= _BV(SUART_IE_BIT);
	} else {
		suart_rxt = temp_tail;
	}



	return c;
}

uint8_t
suart_get_would_block(void)
{
	return (bit_is_set(SUART_IE_REG, SUART_IE_BIT)
			&& suart_rxt == suart_rxh);
}

int
suart_stdio_get(FILE *stream __attribute__((unused)))
{
	return suart_get();
}

int
suart_stdio_put(char c, FILE *stream __attribute__((unused)))
{
	suart_put ((uint8_t) c);
	return 0;
}

