/*
 * Implementation of a half-duplex software UART.
 *
 * Data reception is in an interrupt handler that is busy for the whole
 * duration of the reception of a single byte. Progress in the rest of the
 * program, including other interrupts, is only made between bytes received on
 * the UART. All timing is done through busy-waiting, rather than needing a
 * hardware timer. The rationale for this design is that if reception is done
 * in the main program with busy-waiting, interrupts would need to be turned
 * off as well to get the timing right. Doing all the work in an interrupt
 * handler that triggers on the startbit means the main program is freed from
 * the task of receiving the UART data and can simply look in the circular
 * buffer filled by the interrupt routine.
 *
 * The only supported format is 8n1. For high speeds (brr values lower than
 * 32), the received start bit is ignored. The received stop bit is always
 * ignored.
 *
 * *NOTE* that for very high speeds, differences in interrupt latency can cause
 * timing problems. If the CPU is just executing a CALL or RET instruction or
 * similar, the interrupt delay is so high that the timing is completely wrong
 * or the noise margin is decreased too much. Careful programming could avoid
 * any CALLs or RETs while receiving data.
 *
 * Copyright 2012 Peter Lebbing <peter@digitalbrains.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef FILE_SUART_H
#define FILE_SUART_H

#include <stdio.h>

/********************************************************************
 * The first section contains tunable parameters
 ********************************************************************/

/* Size in bytes of the receive buffer */
#define SUART_BUFSIZE 8
/* Pin number, PORTx, PINx and DDRx register where RxD GPIO-pin is located */
#define SUART_RX_PORT  PORTD
#define SUART_RX_PIN   PIND
#define SUART_RX_DDR   DDRD
#define SUART_RX       PD2
/* Interrupt vector for reception */
#define SUART_vect     INT0_vect
/*
 * Register holding Interrupt Enable bit for reception, and Interrupt Enable
 * bit
 */
#define SUART_IE_REG   EIMSK
#define SUART_IE_BIT   INT0
/* Register holding Interrupt Flag bit for reception, and Interrupt Flag bit */
#define SUART_FLAG_REG EIFR
#define SUART_FLAG_BIT INTF0
/* Pin number, PORTx, PINx and DDRx register where TxD GPIO-pin is located */
#define SUART_TX_PORT  PORTD
#define SUART_TX_DDR   DDRD
#define SUART_TX       PD3

/* Non-zero will install the SUART as stdin/stdout */
#define SUART_AS_STDIO 0

/********************************************************************
 * The next section contains public definitions that are not tunable
 ********************************************************************/

/* Round baudrate to nearest brr value */
#define SUART_BAUD_TO_BRR(x) ((F_CPU / (x) * 2 + 3) / 6 - 3)

#ifndef __ASSEMBLER__

/*
 * Initialise the software UART to the given bitrate
 *
 * brr indicates the desired bitrate according to the following formula:
 *
 * bitrate = F_CPU / (9 + 3 * brr)
 *
 * The range of brr is 1-32766, giving a bitrate divisor in the range of 12
 * to 98307 clock ticks.
 *
 * Warning: The brr is not checked for correctness!
 *
 * You can use SUART_BAUD_TO_BRR to compute the brr for a given baudrate.
 *
 * This routine only enables the used interrupt; any other setup (f.e.,
 * selecting falling edge trigger for INTx) needs to be done beforehand.
 */
extern void
suart_init(uint16_t brr);

/* Disable reception interrupt */
static inline void
suart_disable_rx(void)
{
	SUART_IE_REG &= ~_BV(SUART_IE_BIT);
}

/* Enable reception interrupt */
extern void
suart_enable_rx(void);

/*
 * Get a byte received from the software UART
 *
 * Blocks until a byte is received.
 */
extern uint8_t
suart_get(void);

/*
 * Returns non-zero if suart_get() would block
 */
extern uint8_t
suart_get_would_block(void);

/*
 * stdio-compatible get function
 * 
 * End-Of-File is never returned; it simply waits for more data.
 */
extern int
suart_stdio_get(FILE *stream);

/*
 * Send a byte through the software UART
 *
 * Interrupts are disabled in the timing-critical section.
 */
extern void
suart_put(uint8_t c);

/*
 * stdio-compatible put function
 *
 * Wrapper around suart_put().
 *
 * Always returns 0.
 */
extern int
suart_stdio_put(char c, FILE *stream);

/********************************************************************
 * The rest are internals
 ********************************************************************/

#if (SUART_BUFSIZE & (SUART_BUFSIZE - 1))
#  error "SUART_BUFSIZE needs to be a power of two"
#endif

extern volatile uint8_t
suart_brr_init;

extern volatile uint8_t
suart_brr1;

extern volatile uint8_t
suart_brr2;

extern volatile uint8_t
suart_rxh,
suart_rxt;

extern volatile uint8_t
suart_buf[SUART_BUFSIZE];

#endif /* ndef __ASSEMBLER__ */

/*
 * A set highest bit of suart_brr2 indicates a "small" brr
 */
#define SUART_SMALL_BRR 7

#endif /* ndef FILE_SUART_H */
