; Copyright 2012 Peter Lebbing <peter@digitalbrains.com>
;
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <avr/io.h>
#include "common.h"
#include "suart.h"
#include "asmdelay.inc"

; The timing of the "small brr" reception routine is as follows:
;
; Supposing that the brr value is exactly the desired bitrate, the following 
; tables relate the desired sample moment ("exact") against the moment the
; sample is actually taken ("actual"). Two variations come out, one for even brr
; values, one for odd.
;
; All values are in clockticks.
;
; Period=12 (brr_init=1, brr=1):       Period=15 (brr_init=2, brr=2):
; bit    exact actual diff             bit    exact actual diff
; 0         18     19   +1             0       22.5     22 -0.5
; 1         30     30                  1       37.5     36 -1.5
; 2         42     41   -1             2       52.5     51 -1.5
; 3         54     54                  3       67.5     67 -0.5
; 4         66     66                  4       82.5     82 -0.5
; 5         78     78                  5       97.5     97 -0.5
; 6         90     90                  6      112.5    112 -0.5
; 7        102    101   -1             7      127.5    126 -1.5
; Stop     114    113   -1             Stop   142.5    141 -1.5
; Startn   120    120                  Startn 150      148
;                                  
; Period=18 (brr_init=4, brr=3):       Period=21 (brr_init=5, brr=4):
; bit    exact actual diff             bit    exact actual diff
; 0         27     28   +1             0       31.5     31 -0.5
; 1         45     45                  1       52.5     51 -1.5
; 2         63     62   -1             2       73.5     72 -1.5
; 3         81     81                  3       94.5     94 -0.5
; 4         99     99                  4      115.5    115 -0.5
; 5        117    117                  5      136.5    136 -0.5
; 6        135    135                  6      157.5    157 -0.5
; 7        153    152   -1             7      178.5    177 -1.5
; Stop     171    170   -1             Stop   199.5    198 -1.5
; Startn   180    177                  Startn 210      205
;                                  
; brr odd:                             brr even:
; bit rel.cycle                        bit rel.cycle
; 0          +1                        0        -0.5
; 1                                    1        -1.5
; 2          -1                        2        -1.5
; 3                                    3        -0.5
; 4                                    4        -0.5
; 5                                    5        -0.5
; 6                                    6        -0.5
; 7          -1                        7        -1.5
; Stop       -1                        Stop     -1.5
;
; Note that the stopbit is not actually sampled; but the interrupt flag is
; cleared at the moment listed as "stop", so that any triggering edge after that
; moment is taken as the next startbit.
; "Startn" is the first moment that a triggering edge would result in correct
; timing for the reception of the next bit.
;
; Note on delays: if you have terribly bad luck, the interrupt triggers during
; a CALL or RET or similar instruction. This increases interrupt latency, and
; decreases the noise margin for really small values of brr.
;
; Comments below of the form "x cycles so far" are for brr=1.
;
; stack frame small_brr:
; [return addr]
; data
; SREG
; r28 (YL)
; r29 (YH)
; brrl
; tmp

data = 16
tmp  = 20
brrl = 18
brrh = 19
tbrrl = 20
tbrrh = 21

	.global SUART_vect
SUART_vect:
	push	data
	in	data, IO(SREG)
	push	data
; 12 cycles so far
	push	r28

	lds	data, suart_brr_init
1:
	dec	data
	brne	1b
	sbic	IO(SUART_RX_PIN), SUART_RX
	sbr	data, _BV(0)		; databit 0

; 20 cycles so far

	push	r29
	push	brrl
	lds	brrl, suart_brr1

	mov	r28, brrl
1:
	dec	r28
	brne	1b
	sbic	IO(SUART_RX_PIN), SUART_RX
	sbr	data, _BV(1)		; databit 1

; 31 cycles so far

	lds	r28, suart_brr2
	sbrs	r28, SUART_SMALL_BRR
	rjmp	isr_large_brr		; flag bit for short delay not set

	sbrs	brrl, 0			; Delay an extra cycle if brrl even
	rjmp	1f			; Just a no-op that takes 2 cycles :)

1:
	mov	r28, brrl
1:
	dec	r28
	brne	1b
	sbic	IO(SUART_RX_PIN), SUART_RX
	sbr	data, _BV(2)		; databit 2	

; 42 cycles so far

	push	tmp
	lds	r28, suart_rxh
	inc	r28
	andi	r28, (SUART_BUFSIZE - 1)
	sts	suart_rxh, r28

	mov	tmp, brrl
1:
	dec	tmp
	brne	1b
	sbic	IO(SUART_RX_PIN), SUART_RX
	sbr	data, _BV(3)		; databit 3	

; 55 cycles so far

	lds	r29, SUART_IE_REG
	lds	tmp, suart_rxt
	cp	r28, tmp		; head == tail?
	brne	1f
	cbr	r29, _BV(SUART_IE_BIT)	; Yes, disable interrupt

1:
	mov	tmp, brrl
1:
	dec	tmp
	brne	1b
	sbic	IO(SUART_RX_PIN), SUART_RX
	sbr	data, _BV(4)		; databit 4

; 67 cycles so far

	sts	SUART_IE_REG, r29
	ldi	r29, 0
	subi	r28, lo8(-(suart_buf))
	sbci	r29, hi8(-(suart_buf))	; Y = suart_buf[rxh]

	mov	tmp, brrl
1:
	dec	tmp
	brne	1b
	nop
	nop
	sbic	IO(SUART_RX_PIN), SUART_RX
	sbr	data, _BV(5)		; databit 5

; 79 cycles so far


	mov	tmp, brrl
1:
	dec	tmp
	brne	1b

	ldi	tmp, 2
1:
	dec	tmp
	brne	1b
	nop
	sbic	IO(SUART_RX_PIN), SUART_RX
	sbr	data, _BV(6)		; databit 6

; 91 cycles so far

; The endgame depends on brrl

	nop
	mov	tmp, brrl
	dec	tmp
	breq	brrl_1
	dec	tmp
	breq	brrl_2

; --- BRRL > 2 endgame ---
1:
	dec	tmp
	brne	1b

	pop	tmp

	subi	brrl, 2
.rept 4
	nop
.endr
	sbic	IO(SUART_RX_PIN), SUART_RX
	sbr	data, _BV(7)		; databit 7

	st	Y, data

1:
	dec	brrl
	brne	1b

	pop	brrl
	pop	r29
	pop	r28
	pop	data
	out	IO(SREG), data
	ldi	data, _BV(SUART_FLAG_BIT)
	nop
	nop
	out	IO(SUART_FLAG_REG), data
	pop	data
	reti

; --- BRRL = 1 endgame ---
brrl_1:
	pop	tmp
	pop	brrl
	sbic	IO(SUART_RX_PIN), SUART_RX
	sbr	data, _BV(7)		; databit 7

; 102 cycles so far

	st	Y, data
	pop	r29
	pop	r28
	pop	data
	out	IO(SREG), data
	ldi	data, _BV(SUART_FLAG_BIT)
	out	IO(SUART_FLAG_REG), data
; 113 cycles so far 
	pop	data
	reti

; handler done in 119 cycles

; --- BRRL = 2 endgame ---
brrl_2:
	pop	tmp
	pop	brrl
	nop
	sbic	IO(SUART_RX_PIN), SUART_RX
	sbr	data, _BV(7)		; databit 7

; 126 cycles so far

	st	Y, data
	pop	r29
	pop	r28
	pop	data
	out	IO(SREG), data
	ldi	data, _BV(SUART_FLAG_BIT)
	nop
	nop
	nop
	out	IO(SUART_FLAG_REG), data
; 140 cycles so far
	pop	data
	reti

; handler done in 146 cycles

; stack frame large_brr:
; [return addr]
; data
; SREG
; r28 (YL)
; r29 (YH)
; brrl
; brrh
; tbrrl
; tbrrh

isr_large_brr:
; Currently, brrl = suart_brr1 and r28 = suart_brr2
	mov	r29, brrl		; r29 = suart_brr1
	dec	r29
	lsl	r28
	lsr	r29
	ror	r28			; Y = brr
	lsr	r29
	ror	r28			; Y = 0.5 * brr
	subi	r28, 14			; Correct for constant amount of spent
	sbci	r29, 0			; cycles so far
	sub	r28, brrl		; Already did a wait loop, correct for 
	sbci	r29, 0			; it

	delay16 r29, r28

	sbic	IO(SUART_RX_PIN), SUART_RX
	rjmp	isr_no_start            ; No startbit, interpret as glitch

	push	brrh
	push	tbrrl
	push	tbrrh

	lds	brrl, suart_brr2
	lds	brrh, suart_brr1
	dec	brrh
	lsl	brrl
	lsr	brrh
	ror	brrl			; BRR "computed"
	
	movw	tbrrl, brrl
	subi	tbrrl, 5		; Correct for constant amount of spent
	sbci	tbrrl, 0		; cycles

	ldi	data, 0x80		; Marker bit to see how far we are
	nop
1:
	delay16	tbrrh, tbrrl

	clc
	sbic	IO(SUART_RX_PIN), SUART_RX
	sec
	
	movw	tbrrl, brrl

	ror	data
	brcc	1b

	lds	r28, suart_rxh
	inc	r28
	andi	r28, (SUART_BUFSIZE - 1)

	lds	r29, SUART_IE_REG
	lds	tmp, suart_rxt
	cp	r28, tmp		; head == tail?
	brne	1f
	cbr	r29, _BV(SUART_IE_BIT)	; Yes, disable interrupt
1:
	sts	SUART_IE_REG, r29
	sts	suart_rxh, r28
	ldi	r29, 0
	subi	r28, lo8(-(suart_buf))
	sbci	r29, hi8(-(suart_buf))	; Y = suart_buf[rxh]

	st	Y, data

	subi	brrl, 12		; Put flag clear in the middle of the
	sbci	brrh, 0			; stopbit

	delay16 brrh, brrl

	pop	tbrrh
	pop	tbrrl
	pop	brrh
isr_no_start:
	pop	brrl
	pop	r29
	pop	r28
	pop	data
	out	IO(SREG), data
	ldi	data, _BV(SUART_FLAG_BIT)
	out	IO(SUART_FLAG_REG), data
	pop	data
	reti

; Function:
; void suart_put(uint8_t c)

data = 24
port = 18
sregs = 19
brrl = 26
brrh = 27
tbrrl = 20
tbrrh = 21

	.global suart_put
suart_put:
	lds	brrl, suart_brr2
	sbrs	brrl, SUART_SMALL_BRR
	rjmp	put_large_brr

	lds	brrl, suart_brr1
	ldi	brrh, 0
	rjmp	put_do

put_large_brr:
	lds	brrh, suart_brr1
	dec	brrh
	lsl	brrl
	lsr	brrh
	ror	brrl

put_do:
	clt				; Startbit
	
	movw	tbrrl, brrl

	in	sregs, IO(SREG)
	cli

	in	port, IO(SUART_TX_PORT)

	bld	port, SUART_TX
	out	IO(SUART_TX_PORT), port

put_maybe_stop:
	delay16 tbrrh, tbrrl

	brtc	put_data		; Branch if this was the startbit
	out	IO(SREG), sregs
	ret

put_data:
	sec				; Shift stopbit in data reg
	bst	data, 0
	ror	data

1:
	bld	port, SUART_TX
	out	IO(SUART_TX_PORT), port

	movw	tbrrl, brrl
	delay16	tbrrh, tbrrl

	bst	data, 0
	lsr	data
	brne	1b

; Still need to actually *send* stopbit...	
	movw	tbrrl, brrl
	bld	port, SUART_TX
	out	IO(SUART_TX_PORT), port
	rjmp	put_maybe_stop
